"""
Utilities for interfacing with the
`ETA library <https://github.com/voxel51/eta>`_.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import defaultdict
import warnings

import numpy as np

import eta.core.data as etad
import eta.core.events as etae
import eta.core.frames as etaf
import eta.core.frameutils as etafu
import eta.core.geometry as etag
import eta.core.image as etai
import eta.core.keypoints as etak
import eta.core.learning as etal
import eta.core.objects as etao
import eta.core.polylines as etap
import eta.core.utils as etau
import eta.core.video as etav

import fiftyone.core.labels as fol
import fiftyone.core.models as fom


_IMAGE_MODELS = (
    etal.ImageModel,
    etal.ImageClassifier,
    etal.ObjectDetector,
    etal.ImageSemanticSegmenter,
)

_VIDEO_MODELS = (
    etal.VideoModel,
    etal.VideoClassifier,
    etal.VideoObjectDetector,
    etal.VideoEventDetector,
    etal.VideoSemanticSegmenter,
)


class ETAModelConfig(fom.ModelConfig):
    """Meta-config class that encapsulates the configuration of an
    `eta.core.learning.Model` that is to be run via the :class:`ETAModel`
    wrapper.

    Example::

        import fiftyone.core.models as fom

        model = fom.load_model({
            "type": "fiftyone.utils.eta.ETAModel",
            "config": {
                "type": "eta.detectors.YOLODetector",
                "config": {
                    "model_name": "yolo-v2-coco"
                }
            }
        })

    Args:
        type: the fully-qualified class name of the
            :class:`fiftyone.core.models.Model` subclass, which must be
            :class:`ETAModel` or a subclass of it
        config: a dict containing the ``eta.core.learning.ModelConfig`` for the
            ETA model
    """

    @property
    def confidence_thresh(self):
        """The confidence threshold of the underlying ``eta.core.model.Model``.

        Note that this may not be defined for some models.
        """
        return self.config.confidence_thresh

    @confidence_thresh.setter
    def confidence_thresh(self, confidence_thresh):
        self.config.confidence_thresh = confidence_thresh


class ETAModel(fom.Model, fom.EmbeddingsMixin, fom.LogitsMixin):
    """Wrapper for running an ``eta.core.learning.Model`` model.

    Args:
        config: an :class:`ETAModelConfig`
    """

    def __init__(self, config, _model=None):
        if _model is None:
            _model = config.build()  # build the ETA model

        self.config = config
        self._model = _model
        fom.LogitsMixin.__init__(self)

    def __enter__(self):
        self._model.__enter__()
        return self

    def __exit__(self, *args):
        self._model.__exit__(*args)

    @property
    def media_type(self):
        if isinstance(self._model, _IMAGE_MODELS):
            return "image"

        if isinstance(self._model, _VIDEO_MODELS):
            return "video"

        return None

    @property
    def ragged_batches(self):
        try:
            return self._model.ragged_batches
        except AttributeError:
            return True

    @property
    def transforms(self):
        try:
            return self._model.transforms
        except AttributeError:
            return None

    @property
    def preprocess(self):
        try:
            return self._model.preprocess
        except AttributeError:
            return False

    @preprocess.setter
    def preprocess(self, value):
        try:
            self._model.preprocess = value
        except AttributeError:
            pass

    @property
    def has_logits(self):
        return (
            isinstance(self._model, etal.ExposesProbabilities)
            and isinstance(self._model, etal.Classifier)
            and self._model.exposes_probabilities
        )

    @property
    def has_embeddings(self):
        return (
            isinstance(self._model, etal.ExposesFeatures)
            and isinstance(self._model, etal.Classifier)
            and self._model.exposes_features
        )

    def _ensure_embeddings(self):
        if not self.has_embeddings:
            raise ValueError("This model instance does not expose embeddings")

    def get_embeddings(self):
        self._ensure_embeddings()
        embeddings = self._model.get_features()
        embeddings = _squeeze_extra_unit_dims(embeddings)
        return embeddings.astype(float, copy=False)

    def embed(self, arg):
        self._ensure_embeddings()
        self.predict(arg)
        return self.get_embeddings()

    def embed_all(self, args):
        self._ensure_embeddings()

        if isinstance(self._model, etal.ImageClassifier):
            self._model.predict_all(args)
            return self.get_embeddings()

        if isinstance(self._model, etal.ObjectDetector):
            self._model.detect_all(args)
            return self.get_embeddings()

        if isinstance(self._model, etal.ImageSemanticSegmenter):
            self._model.segment_all(args)
            return self.get_embeddings()

        return np.concatenate(tuple(self.embed(arg) for arg in args))

    def predict(self, arg):
        if isinstance(self._model, etal.ImageClassifier):
            eta_labels = self._model.predict(arg)
        elif isinstance(self._model, etal.VideoFramesClassifier):
            eta_labels = self._model.predict(arg)
        elif isinstance(self._model, etal.VideoClassifier):
            eta_labels = self._model.predict(arg)
        elif isinstance(self._model, etal.Classifier):
            eta_labels = self._model.predict(arg)
        elif isinstance(self._model, etal.ObjectDetector):
            eta_labels = self._model.detect(arg)
        elif isinstance(self._model, etal.VideoFramesObjectDetector):
            eta_labels = self._model.detect(arg)
        elif isinstance(self._model, etal.VideoObjectDetector):
            eta_labels = self._model.detect(arg)
        elif isinstance(self._model, etal.Detector):
            eta_labels = self._model.detect(arg)
        elif isinstance(self._model, etal.ImageSemanticSegmenter):
            eta_labels = self._model.segment(arg)
        elif isinstance(self._model, etal.VideoSemanticSegmenter):
            eta_labels = self._model.segment(arg)
        elif isinstance(self._model, etal.SemanticSegmenter):
            eta_labels = self._model.segment(arg)
        elif isinstance(self._model, etal.ImageModel):
            eta_labels = self._model.process(arg)
        elif isinstance(self._model, etal.VideoModel):
            eta_labels = self._model.process(arg)
        else:
            raise ValueError(
                "Unsupported model type '%s'" % self._model.__class__
            )

        eta_labels = self._parse_predictions(eta_labels)

        label = _from_eta_labels(eta_labels)

        if self.has_logits and self.store_logits:
            # num_preds x num_classes
            logits = np.log(self._model.get_probabilities()[0])
            _add_logits(label, logits)

        return label

    def predict_all(self, args):
        if isinstance(self._model, etal.ImageClassifier):
            eta_labels_batch = self._model.predict_all(args)
        elif isinstance(self._model, etal.ObjectDetector):
            eta_labels_batch = self._model.detect_all(args)
        elif isinstance(self._model, etal.ImageSemanticSegmenter):
            eta_labels_batch = self._model.segment_all(args)
        else:
            return [self.predict(arg) for arg in args]

        eta_labels_batch = self._parse_predictions(eta_labels_batch)

        labels = [_from_eta_labels(el) for el in eta_labels_batch]

        if self.has_logits and self.store_logits:
            # num_images x num_preds x num_classes
            logits = np.log(self._model.get_probabilities())

            for label, _logits in zip(labels, logits):
                _add_logits(label, _logits)

        return labels

    def _parse_predictions(self, eta_labels_or_batch):
        if (
            not isinstance(self._model, etal.Classifier)
            or self._model.is_multilabel
        ):
            return eta_labels_or_batch

        if isinstance(eta_labels_or_batch, list):
            return [
                attrs[0] if attrs else None for attrs in eta_labels_or_batch
            ]

        if not eta_labels_or_batch:
            return None

        return eta_labels_or_batch[0]

    @classmethod
    def from_eta_model(cls, model):
        """Builds an :class:`ETAModel` for running the provided
        ``eta.core.learning.Model`` instance.

        Args:
            model: an ``eta.core.learning.Model`` instance

        Returns:
            an :class:`ETAModel`
        """
        return cls(model.config, _model=model)


def from_image_labels(
    image_labels_or_path,
    prefix=None,
    labels_dict=None,
    multilabel=False,
    skip_non_categorical=False,
):
    """Loads the ``eta.core.image.ImageLabels`` or
    ``eta.core.frames.FrameLabels`` into a dictionary of labels.

    Provide ``labels_dict`` if you want to customize which components of
    the labels are expanded. Otherwise, all labels are expanded as
    explained below.

    If ``multilabel`` is False, frame attributes will be stored in separate
    :class:`Classification` fields with names ``prefix + attr.name``.

    If ``multilabel`` if True, all frame attributes will be stored in a
    :class:`Classifications` field called ``prefix + "attributes"``.

    Objects are expanded into fields with names ``prefix + obj.name``, or
    ``prefix + "detections"`` for objects that do not have their ``name``
    field populated.

    Polylines are expanded into fields with names
    ``prefix + polyline.name``, or ``prefix + "polylines"`` for polylines
    that do not have their ``name`` field populated.

    Keypoints are expanded into fields with names
    ``prefix + keypoints.name``, or ``prefix + "keypoints"`` for keypoints
    that do not have their ``name`` field populated.

    Segmentation masks are expanded into a field with name ``prefix + "mask"``.

    Args:
        image_labels_or_path: can be a ``eta.core.image.ImageLabels`` instance,
            a ``eta.core.frames.FrameLabels`` instance, a serialized dict
            representation of either, or the path to either on disk
        prefix (None): a string prefix to prepend to each field name in the
            output dict
        labels_dict (None): a dictionary mapping names of labels to keys to
            assign them in the output dictionary
        multilabel (False): whether to store attributes in a single
            :class:`Classifications` instance
        skip_non_categorical (False): whether to skip non-categorical
            attributes (True) or cast them to strings (False)

    Returns:
        a dict mapping names to :class:`fiftyone.core.labels.Label` instances
    """
    if etau.is_str(image_labels_or_path):
        frame_labels = etaf.FrameLabels.from_json(image_labels_or_path)
    elif isinstance(image_labels_or_path, dict):
        frame_labels = etaf.FrameLabels.from_dict(image_labels_or_path)
    else:
        frame_labels = image_labels_or_path

    if frame_labels is None:
        return None

    if labels_dict is not None:
        return _expand_with_labels_dict(
            frame_labels, labels_dict, multilabel, skip_non_categorical
        )

    return _expand_with_prefix(
        frame_labels, prefix, multilabel, skip_non_categorical
    )


def to_image_labels(labels, warn_unsupported=True):
    """Converts the image label(s) to ``eta.core.image.ImageLabels`` format.

    Args:
        labels: a :class:`fiftyone.core.labels.Label` instance or a dict
            mapping names to :class:`fiftyone.core.labels.Label` instances
        warn_unsupported (True): whether to issue warnings if unsupported label
            values are encountered

    Returns:
        an ``eta.core.image.ImageLabels`` instance
    """
    image_labels = etai.ImageLabels()

    if labels is None:
        return image_labels

    if not isinstance(labels, dict):
        labels = {"labels": labels}

    _add_frame_labels(image_labels, labels, warn_unsupported=warn_unsupported)

    return image_labels


def from_video_labels(
    video_labels_or_path,
    prefix=None,
    labels_dict=None,
    frame_labels_dict=None,
    multilabel=False,
    skip_non_categorical=False,
):
    """Loads the ``eta.core.video.VideoLabels`` into a frame labels dictionary.

    Args:
        video_labels_or_path: can be a ``eta.core.video.VideoLabels`` instance,
            a serialized dict representation of one, or the path to one on disk
        prefix (None): a string prefix to prepend to each label name in the
            expanded sample/frame label dictionaries
        labels_dict (None): a dictionary mapping names of attributes/objects
            in the sample labels to field names into which to expand them. By
            default, all sample labels are loaded
        frame_labels_dict (None): a dictionary mapping names of
            attributes/objects in the frame labels to field names into which to
            expand them. By default, all frame labels are loaded
        multilabel (False): whether to store attributes in a single
            :class:`fiftyone.core.labels.Classifications` instance
        skip_non_categorical (False): whether to skip non-categorical
            attributes (True) or cast them to strings (False)

    Returns:
        a tuple of

        -   **label**: a dict mapping sample field names to
            :class:`fiftyone.core.labels.Label` instances
        -   **frames**: a dict mapping frame numbers to dicts that map label
            fields to :class:`fiftyone.core.labels.Label` instances
    """
    if etau.is_str(video_labels_or_path):
        video_labels = etav.VideoLabels.from_json(video_labels_or_path)
    elif isinstance(video_labels_or_path, dict):
        video_labels = etav.VideoLabels.from_dict(video_labels_or_path)
    else:
        video_labels = video_labels_or_path

    if video_labels is None:
        return None, None

    # Video labels
    if (
        video_labels.has_video_attributes or video_labels.has_video_events
    ) and (labels_dict is None or labels_dict):
        if labels_dict is not None:
            label = _expand_with_labels_dict(
                video_labels, labels_dict, multilabel, skip_non_categorical
            )
        else:
            label = _expand_with_prefix(
                video_labels, prefix, multilabel, skip_non_categorical
            )
    else:
        label = None

    # Frame labels
    if video_labels.frames and (
        frame_labels_dict is None or frame_labels_dict
    ):
        frames = {}
        for frame_number in video_labels:
            frames[frame_number] = from_image_labels(
                video_labels[frame_number],
                prefix=prefix,
                labels_dict=frame_labels_dict,
                multilabel=multilabel,
                skip_non_categorical=skip_non_categorical,
            )
    else:
        frames = None

    return label, frames


def to_video_labels(
    label=None, frames=None, support=None, warn_unsupported=True
):
    """Converts the given labels to ``eta.core.video.VideoLabels`` format.

    Args:
        label (None): video-level labels provided as a
            :class:`fiftyone.core.labels.Label` instance or dict mapping field
            names to :class:`fiftyone.core.labels.Label` instances
        frames (None): frame-level labels provided as a dict mapping frame
            numbers to dicts mapping field names to
            :class:`fiftyone.core.labels.Label` instances
        support (None): an optional ``[first, last]`` support to store on the
            returned labels
        warn_unsupported (True): whether to issue warnings if unsupported label
            values are encountered

    Returns:
        a ``eta.core.video.VideoLabels``
    """
    if support is not None:
        support = etafu.FrameRanges.build_simple(*support)

    video_labels = etav.VideoLabels(support=support)

    # Video labels
    if label is not None:
        if not isinstance(label, dict):
            label = {"labels": label}

        _add_video_labels(
            video_labels, label, warn_unsupported=warn_unsupported
        )

    # Frame labels
    if frames is not None:
        for frame_number, frame in frames.items():
            frame_labels = etav.VideoFrameLabels(frame_number)

            _add_frame_labels(
                frame_labels, frame, warn_unsupported=warn_unsupported
            )

            video_labels[frame_number] = frame_labels

    return video_labels


def to_attribute(classification, name=None):
    """Returns an ``eta.core.data.Attribute`` representation of the
    :class:`fiftyone.core.labels.Classification`.

    Args:
        classification: a :class:`fiftyone.core.labels.Classification`
        name (None): the name of the label field

    Returns:
        a ``eta.core.data.CategoricalAttribute``
    """
    return etad.CategoricalAttribute(
        name,
        classification.label,
        confidence=classification.confidence,
        tags=classification.tags,
    )


def from_attribute(attr):
    """Creates a :class:`fiftyone.core.labels.Classification` from an
    ``eta.core.data.Attribute``.

    The attribute value is cast to a string, if necessary.

    Args:
        attr: an ``eta.core.data.Attribute``

    Returns:
        a :class:`fiftyone.core.labels.Classification`
    """
    classification = fol.Classification(label=str(attr.value))

    try:
        classification.confidence = attr.confidence
    except:
        pass

    return classification


def from_attributes(attrs, skip_non_categorical=False):
    """Creates a :class:`fiftyone.core.labels.Classifications` from a list of
    attributes.

    Args:
        attrs: an iterable of ``eta.core.data.Attribute`` instances
        skip_non_categorical (False): whether to skip non-categorical
            attributes (True) or cast all attribute values to strings
            (False)

    Returns:
        a :class:`fiftyone.core.labels.Classifications`
    """
    classifications = []
    for attr in attrs:
        if skip_non_categorical and not etau.is_str(attr.value):
            continue

        classifications.append(from_attribute(attr))

    return fol.Classifications(classifications=classifications)


def to_detected_object(detection, name=None, extra_attrs=True):
    """Returns an ``eta.core.objects.DetectedObject`` representation of the
    given :class:`fiftyone.core.labels.Detection`.

    Args:
        detection: a :class:`fiftyone.core.labels.Detection`
        name (None): the name of the label field
        extra_attrs (True): whether to include custom attributes in the
            conversion

    Returns:
        an ``eta.core.objects.DetectedObject``
    """
    label = detection.label
    index = detection.index

    # pylint: disable=unpacking-non-sequence
    tlx, tly, w, h = detection.bounding_box
    brx = tlx + w
    bry = tly + h
    bounding_box = etag.BoundingBox.from_coords(tlx, tly, brx, bry)

    mask = detection.mask
    confidence = detection.confidence

    attrs = _to_eta_attributes(detection, extra_attrs=extra_attrs)

    return etao.DetectedObject(
        label=label,
        index=index,
        bounding_box=bounding_box,
        mask=mask,
        confidence=confidence,
        name=name,
        attrs=attrs,
        tags=detection.tags,
    )


def from_detected_object(dobj):
    """Creates a :class:`fiftyone.core.labels.Detection` from an
    ``eta.core.objects.DetectedObject``.

    Args:
        dobj: a ``eta.core.objects.DetectedObject``

    Returns:
        a :class:`fiftyone.core.labels.Detection`
    """
    xtl, ytl, xbr, ybr = dobj.bounding_box.to_coords()
    bounding_box = [xtl, ytl, (xbr - xtl), (ybr - ytl)]

    attributes = _from_eta_attributes(dobj.attrs)

    return fol.Detection(
        label=dobj.label,
        bounding_box=bounding_box,
        confidence=dobj.confidence,
        index=dobj.index,
        mask=dobj.mask,
        tags=dobj.tags,
        **attributes,
    )


def from_detected_objects(objects):
    """Creates a :class:`fiftyone.core.labels.Detections` from an
    ``eta.core.objects.DetectedObjectContainer``.

    Args:
        objects: a ``eta.core.objects.DetectedObjectContainer``

    Returns:
        a :class:`fiftyone.core.labels.Detections`
    """
    return fol.Detections(
        detections=[from_detected_object(dobj) for dobj in objects]
    )


def to_polyline(polyline, name=None, extra_attrs=True):
    """Returns an ``eta.core.polylines.Polyline`` representation of the given
    :class:`fiftyone.core.labels.Polyline`.

    Args:
        polyline: a :class:`fiftyone.core.labels.Polyline`
        name (None): the name of the label field
        extra_attrs (True): whether to include custom attributes in the
            conversion

    Returns:
        an ``eta.core.polylines.Polyline``
    """
    attrs = _to_eta_attributes(polyline, extra_attrs=extra_attrs)

    return etap.Polyline(
        label=polyline.label,
        confidence=polyline.confidence,
        index=polyline.index,
        name=name,
        points=polyline.points,
        closed=polyline.closed,
        filled=polyline.filled,
        attrs=attrs,
        tags=polyline.tags,
    )


def from_polyline(polyline):
    """Creates a :class:`fiftyone.core.labels.Polyline` from an
    ``eta.core.polylines.Polyline``.

    Args:
        polyline: an ``eta.core.polylines.Polyline``

    Returns:
        a :class:`fiftyone.core.labels.Polyline`
    """
    attributes = _from_eta_attributes(polyline.attrs)

    return fol.Polyline(
        label=polyline.label,
        points=polyline.points,
        confidence=polyline.confidence,
        index=polyline.index,
        closed=polyline.closed,
        filled=polyline.filled,
        tags=polyline.tags,
        **attributes,
    )


def from_polylines(polylines):
    """Creates a :class:`fiftyone.core.labels.Polylines` from an
    ``eta.core.polylines.PolylineContainer``.

    Args:
        polylines: an ``eta.core.polylines.PolylineContainer``

    Returns:
        a :class:`fiftyone.core.labels.Polylines`
    """
    return fol.Polylines(polylines=[from_polyline(p) for p in polylines])


def to_keypoints(keypoint, name=None, extra_attrs=True):
    """Returns an ``eta.core.keypoints.Keypoints`` representation of the given
    :class:`fiftyone.core.labels.Keypoint`.

    Args:
        keypoint: a :class:`fiftyone.core.labels.Keypoint`
        name (None): the name of the label field
        extra_attrs (True): whether to include custom attributes in the
            conversion

    Returns:
        an ``eta.core.keypoints.Keypoints``
    """
    attrs = _to_eta_attributes(keypoint, extra_attrs=extra_attrs)

    return etak.Keypoints(
        name=name,
        label=keypoint.label,
        index=keypoint.index,
        points=keypoint.points,
        confidence=keypoint.confidence,
        attrs=attrs,
        tags=keypoint.tags,
    )


def from_keypoint(keypoints):
    """Creates a :class:`fiftyone.core.labels.Keypoint` from an
    ``eta.core.keypoints.Keypoints``.

    Args:
        keypoints: an ``eta.core.keypoints.Keypoints``

    Returns:
        a :class:`fiftyone.core.labels.Keypoint`
    """
    attributes = _from_eta_attributes(keypoints.attrs)

    return fol.Keypoint(
        label=keypoints.label,
        points=keypoints.points,
        confidence=keypoints.confidence,
        index=keypoints.index,
        tags=keypoints.tags,
        **attributes,
    )


def from_keypoints(keypoints):
    """Creates a :class:`fiftyone.core.labels.Keypoints` from an
    ``eta.core.keypoints.KeypointsContainer``.

    Args:
        keypoints: an ``eta.core.keypoints.KeypointsContainer``

    Returns:
        a :class:`fiftyone.core.labels.Keypoints`
    """
    return fol.Keypoints(keypoints=[from_keypoint(k) for k in keypoints])


def to_video_event(temporal_detection, name=None, extra_attrs=True):
    """Returns an ``eta.core.events.VideoEvent`` representation of the given
    :class:`fiftyone.core.labels.TemporalDetection`.

    Args:
        temporal_detection: a :class:`fiftyone.core.labels.TemporalDetection`
        name (None): the name of the label field
        extra_attrs (True): whether to include custom attributes in the
            conversion

    Returns:
        an ``eta.core.events.VideoEvent``
    """
    support = etafu.FrameRanges.build_simple(*temporal_detection.support)
    attrs = _to_eta_attributes(temporal_detection, extra_attrs=extra_attrs)

    return etae.VideoEvent(
        label=temporal_detection.label,
        confidence=temporal_detection.confidence,
        name=name,
        support=support,
        attrs=attrs,
        tags=temporal_detection.tags,
    )


def from_video_event(video_event):
    """Creates a :class:`fiftyone.core.labels.TemporalDetection` from an
    ``eta.core.events.VideoEvent``.

    Args:
        video_event: an ``eta.core.events.VideoEvent``

    Returns:
        a :class:`fiftyone.core.labels.TemporalDetection`
    """
    if video_event.support:
        support = list(video_event.support.limits)
    else:
        support = None

    attributes = _from_eta_attributes(video_event.attrs)

    return fol.TemporalDetection(
        label=video_event.label,
        support=support,
        confidence=video_event.confidence,
        tags=video_event.tags,
        **attributes,
    )


def from_video_events(video_events):
    """Creates a :class:`fiftyone.core.labels.TemporalDetections` from an
    ``eta.core.events.VideoEventContainer``.

    Args:
        video_events: an ``eta.core.events.VideoEventContainer``

    Returns:
        a :class:`fiftyone.core.labels.TemporalDetections`
    """
    return fol.TemporalDetections(
        detections=[from_video_event(e) for e in video_events]
    )


def _add_frame_labels(frame_labels, labels, warn_unsupported=True):
    for name, label in labels.items():
        if isinstance(label, fol.Classification):
            frame_labels.add_attribute(to_attribute(label, name=name))
        elif isinstance(label, fol.Classifications):
            for classification in label.classifications:
                attr = to_attribute(classification, name=name)
                frame_labels.add_attribute(attr)
        elif isinstance(label, fol.Detection):
            frame_labels.add_object(to_detected_object(label, name=name))
        elif isinstance(label, fol.Detections):
            for detection in label.detections:
                dobj = to_detected_object(detection, name=name)
                frame_labels.add_object(dobj)
        elif isinstance(label, fol.Polyline):
            frame_labels.add_polyline(to_polyline(label, name=name))
        elif isinstance(label, fol.Polylines):
            for polyline in label.polylines:
                poly = to_polyline(polyline, name=name)
                frame_labels.add_polyline(poly)
        elif isinstance(label, fol.Keypoint):
            frame_labels.add_keypoints(to_keypoints(label, name=name))
        elif isinstance(label, fol.Keypoints):
            for keypoint in label.keypoints:
                kp = to_keypoints(keypoint, name=name)
                frame_labels.add_keypoints(kp)
        elif isinstance(label, fol.Segmentation):
            frame_labels.mask = label.mask
            frame_labels.tags.extend(label.tags)
        elif warn_unsupported and label is not None:
            msg = "Ignoring unsupported label type '%s'" % label.__class__
            warnings.warn(msg)

    return frame_labels


def _add_video_labels(video_labels, labels, warn_unsupported=True):
    for name, label in labels.items():
        if isinstance(label, fol.Classification):
            video_labels.add_video_attribute(to_attribute(label, name=name))
        elif isinstance(label, fol.Classifications):
            for classification in label.classifications:
                attr = to_attribute(classification, name=name)
                video_labels.add_video_attribute(attr)
        elif isinstance(label, fol.TemporalDetection):
            video_labels.add_event(to_video_event(label, name=name))
        elif isinstance(label, fol.TemporalDetections):
            for detection in label.detections:
                event = to_video_event(detection, name=name)
                video_labels.add_event(event)
        elif warn_unsupported and label is not None:
            msg = "Ignoring unsupported label type '%s'" % label.__class__
            warnings.warn(msg)


def _from_eta_labels(eta_labels):
    if isinstance(eta_labels, etad.AttributeContainer):
        label = from_attributes(eta_labels)
    elif isinstance(eta_labels, etad.Attribute):
        label = from_attribute(eta_labels)
    elif isinstance(eta_labels, etao.DetectedObjectContainer):
        label = from_detected_objects(eta_labels)
    elif isinstance(eta_labels, etao.DetectedObject):
        label = from_detected_object(eta_labels)
    elif isinstance(eta_labels, etap.PolylineContainer):
        label = from_polylines(eta_labels)
    elif isinstance(eta_labels, etap.Polyline):
        label = from_polyline(eta_labels)
    elif isinstance(eta_labels, etak.KeypointsContainer):
        label = from_keypoints(eta_labels)
    elif isinstance(eta_labels, etak.Keypoints):
        label = from_keypoint(eta_labels)
    elif isinstance(eta_labels, etae.VideoEventContainer):
        label = from_video_events(eta_labels)
    elif isinstance(eta_labels, etae.VideoEvent):
        label = from_video_event(eta_labels)
    elif isinstance(eta_labels, etav.VideoLabels):
        _, label = from_video_labels(eta_labels)  # only frame labels
    elif isinstance(eta_labels, etaf.FrameLabels):
        label = from_image_labels(eta_labels)
    elif isinstance(eta_labels, np.ndarray):
        label = fol.Segmentation(mask=eta_labels)
    elif eta_labels is None:
        label = None
    else:
        raise ValueError(
            "Unsupported ETA label type '%s'" % eta_labels.__class__
        )

    return label


def _from_eta_attributes(attrs):
    return {a.name: a.value for a in attrs}


def _to_eta_attributes(label, extra_attrs=True, warn_unsupported=True):
    attrs = etad.AttributeContainer()

    if not extra_attrs:
        return attrs

    for name, value in label.iter_attributes():
        if etau.is_str(value):
            attrs.add(etad.CategoricalAttribute(name, value))
        elif etau.is_numeric(value):
            attrs.add(etad.NumericAttribute(name, value))
        elif isinstance(value, bool):
            attrs.add(etad.BooleanAttribute(name, value))
        elif warn_unsupported and value is not None:
            msg = "Ignoring unsupported attribute type '%s'" % type(value)
            warnings.warn(msg)

    return attrs


def _expand_with_prefix(
    video_or_frame_labels, prefix, multilabel, skip_non_categorical
):
    if prefix is None:
        prefix = ""

    labels = {}

    #
    # Classifications (both)
    #

    if multilabel:
        # Store frame attributes as multilabels
        labels[prefix + "attributes"] = from_attributes(
            video_or_frame_labels.attrs,
            skip_non_categorical=skip_non_categorical,
        )
    else:
        # Store each frame attribute separately
        for attr in video_or_frame_labels.attrs:
            if skip_non_categorical and not etau.is_str(attr.value):
                continue

            labels[prefix + attr.name] = from_attribute(attr)

    #
    # Temporal detections (video labels only)
    #

    if isinstance(video_or_frame_labels, etav.VideoLabels):
        events_map = defaultdict(etae.VideoEventContainer)

        for event in video_or_frame_labels.events:
            events_map[prefix + (event.name or "events")].add(event)

        for name, events in events_map.items():
            labels[name] = from_video_events(events)

        return labels

    #
    # Detections (frame labels only)
    #

    objects_map = defaultdict(etao.DetectedObjectContainer)

    for dobj in video_or_frame_labels.objects:
        objects_map[prefix + (dobj.name or "detections")].add(dobj)

    for name, objects in objects_map.items():
        labels[name] = from_detected_objects(objects)

    #
    # Polylines (frame labels only)
    #

    polylines_map = defaultdict(etap.PolylineContainer)

    for polyline in video_or_frame_labels.polylines:
        polylines_map[prefix + (polyline.name or "polylines")].add(polyline)

    for name, polylines in polylines_map.items():
        labels[name] = from_polylines(polylines)

    #
    # Keypoints (frame labels only)
    #

    keypoints_map = defaultdict(etak.KeypointsContainer)

    for keypoints in video_or_frame_labels.keypoints:
        keypoints_map[prefix + (keypoints.name or "keypoints")].add(keypoints)

    for name, keypoints in keypoints_map.items():
        labels[name] = from_keypoints(keypoints)

    #
    # Segmentations (frame labels only)
    #

    if video_or_frame_labels.has_mask:
        labels[prefix + "mask"] = fol.Segmentation(
            mask=video_or_frame_labels.mask
        )

    return labels


def _expand_with_labels_dict(
    video_or_frame_labels, labels_dict, multilabel, skip_non_categorical
):
    labels = {}

    #
    # Classifications (both)
    #

    if multilabel:
        # Store frame attributes as multilabels
        attrs_map = defaultdict(etad.AttributeContainer)
        for attr in video_or_frame_labels.attrs:
            if attr.name not in labels_dict:
                continue

            attrs_map[labels_dict[attr.name]].add(attr)

        for name, attrs in attrs_map.items():
            labels[name] = from_attributes(
                attrs, skip_non_categorical=skip_non_categorical
            )
    else:
        # Store each frame attribute separately
        for attr in video_or_frame_labels.attrs:
            if skip_non_categorical and not etau.is_str(attr.value):
                continue

            if attr.name not in labels_dict:
                continue

            labels[labels_dict[attr.name]] = from_attribute(attr)

    #
    # Temporal detections (video labels only)
    #

    if isinstance(video_or_frame_labels, etav.VideoLabels):
        events_map = defaultdict(etae.VideoEventContainer)

        for event in video_or_frame_labels.events:
            if event.name not in labels_dict:
                continue

            events_map[labels_dict[event.name]].add(event)

        for name, events in events_map.items():
            labels[name] = from_video_events(events)

        return labels

    #
    # Detections (frame labels only)
    #

    objects_map = defaultdict(etao.DetectedObjectContainer)

    for dobj in video_or_frame_labels.objects:
        if dobj.name not in labels_dict:
            continue

        objects_map[labels_dict[dobj.name]].add(dobj)

    for name, objects in objects_map.items():
        labels[name] = from_detected_objects(objects)

    #
    # Polylines (frame labels only)
    #

    polylines_map = defaultdict(etap.PolylineContainer)

    for polyline in video_or_frame_labels.polylines:
        if polyline.name not in labels_dict:
            continue

        polylines_map[labels_dict[polyline.name]].add(polyline)

    for name, polylines in polylines_map.items():
        labels[name] = from_polylines(polylines)

    #
    # Keypoints (frame labels only)
    #

    keypoints_map = defaultdict(etak.KeypointsContainer)

    for keypoints in video_or_frame_labels.keypoints:
        if keypoints.name not in labels_dict:
            continue

        keypoints_map[labels_dict[keypoints.name]].add(keypoints)

    for name, keypoints in keypoints_map.items():
        labels[name] = from_keypoints(keypoints)

    #
    # Segmentations (frame labels only)
    #

    if video_or_frame_labels.has_mask and "mask" in labels_dict:
        labels["mask"] = fol.Segmentation(mask=video_or_frame_labels.mask)

    return labels


def _squeeze_extra_unit_dims(embeddings):
    dims = embeddings.shape[1:]
    extra_axes = tuple(ax for ax, dim in enumerate(dims, 1) if dim == 1)

    if len(extra_axes) == len(dims):
        extra_axes = extra_axes[1:]

    if extra_axes:
        return np.squeeze(embeddings, axis=extra_axes)

    return embeddings


def _add_logits(label, logits):
    if isinstance(label, fol.Classification):
        label.logits = logits[0]
    elif isinstance(label, fol.Classifications):
        for c, l in zip(label.classifications, logits):
            c.logits = l
    elif label is not None:
        msg = "Cannot store logits on label type '%s'" % label.__class__
        warnings.warn(msg)



"""
Random sampling utilities.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import numpy as np


def random_split(sample_collection, split_fracs, seed=None):
    """Generates a random partition of the samples in the collection according
    to the specified split fractions.

    Example::

        import fiftyone as fo
        import fiftyone.utils.random as four
        import fiftyone.zoo as foz

        # A dataset with `ground_truth` detections and no tags
        dataset = (
            foz.load_zoo_dataset("quickstart")
            .select_fields("ground_truth")
            .set_field("tags", [])
        ).clone()

        #
        # Generate a random sample and encode results via tags
        #

        four.random_split(dataset, {"train": 0.7, "test": 0.2, "val": 0.1})

        print(dataset.count_sample_tags())
        # {'train': 140, 'test': 40, 'val': 20}

        #
        # Generate a random sample in-memory
        #

        view1, view2 = four.random_split(dataset, [0.5, 0.5])

        assert len(view1) + len(view2) == len(dataset)
        assert set(view1.values("id")).isdisjoint(set(view2.values("id")))

    Args:
        sample_collection: a
            :class:`fiftyone.core.collections.SampleCollection`
        split_fracs: can be either of the following:

            -   a dict mapping tag strings to split fractions in ``[0, 1]``. In
                this case, the partition is denoted by tagging each sample with
                its assigned split
            -   a list of split fractions in ``[0, 1]``. In this case, a
                corresponding list of :class:`fiftyone.core.view.DatasetView`
                instances containing the partition is returned

            In either case, the split fractions are normalized so that they sum
            to 1, if necessary
        seed (None): an optional random seed

    Returns:
        one of the following

        -   ``None``, if ``split_fracs`` is a dict
        -   a tuple of :class:`fiftyone.core.view.DatasetView` instances, if
            ``split_fracs`` is a list
    """
    use_tags = isinstance(split_fracs, dict)

    if use_tags:
        tags, fracs = zip(*split_fracs.items())
    else:
        fracs = split_fracs

    fracs = np.cumsum(fracs)
    alpha = len(sample_collection) / fracs[-1]
    threshs = np.round(alpha * fracs).astype(int)

    sample_ids = np.array(sample_collection.values("id"))
    rs = np.random.RandomState(seed=seed)  # pylint: disable=no-member
    rs.shuffle(sample_ids)

    split_ids = np.split(sample_ids, threshs[:-1])

    if use_tags:
        for ids, tag in zip(split_ids, tags):
            sample_collection.select(ids).tag_samples(tag)

        return

    return tuple(sample_collection.select(ids) for ids in split_ids)


def weighted_sample(
    sample_collection,
    k,
    weights,
    tag=None,
    exact=True,
    seed=None,
):
    """Generates a random sample of size ``k`` from the given collection such
    that the probability of selecting each sample is proportional to the given
    per-sample weights.

    Example::

        import fiftyone as fo
        import fiftyone.utils.random as four
        import fiftyone.zoo as foz
        from fiftyone import ViewField as F

        dataset = foz.load_zoo_dataset("cifar10", split="train")

        # Sample proportional to label length
        weights = dataset.values(F("ground_truth.label").strlen())
        sample = four.weighted_sample(dataset, 10000, weights)

        # Plot results
        plot = fo.CategoricalHistogram(
            "ground_truth.label",
            order=lambda kv: -len(kv[0]),  # order by label length
            init_view=sample,
        )
        plot.show()

    Args:
        sample_collection: a :class:`fiftyone.core.collections.SampleCollection`
        k: the number of samples to select
        weights: an array of per-sample weights
        tag (None): an optional sample tag to use to encode the results
        exact (True): whether to tag exactly ``k`` samples (True) or sample so
            that the expected number of samples is ``k`` (False)
        seed (None): an optional random seed to use

    Returns:
        a :class:`fiftyone.core.view.DatasetView` containing the sample
    """
    weights = np.asarray(weights)
    probs = (k / sum(weights)) * weights

    rs = np.random.RandomState(seed=seed)  # pylint: disable=no-member
    rands = rs.rand(len(sample_collection))

    if exact:
        inds = np.argpartition(rands - probs, k)[:k]
    else:
        inds = rands < probs

    ids = np.array(sample_collection.values("id"))
    select_ids = ids[inds]

    view = sample_collection[select_ids]

    if tag is not None:
        view.tag_samples(tag)
        view = sample_collection.match_tags(tag)

    return view


def balanced_sample(
    sample_collection,
    k,
    path,
    tag=None,
    exact=True,
    seed=None,
):
    """Generates a random sample of size ``k`` from the given collection such
    that the expected histogram of ``path`` values in the sample is uniform.

    Example::

        import fiftyone as fo
        import fiftyone.utils.random as four
        import fiftyone.zoo as foz
        from fiftyone import ViewField as F

        dataset = foz.load_zoo_dataset("cifar10", split="train")

        # Sample proportional to label length
        weights = dataset.values(F("ground_truth.label").strlen())
        view1 = four.weighted_sample(dataset, 10000, weights)

        # Now take a balanced sample from this unbalanced sample
        view2 = four.balanced_sample(view1, 2000, "ground_truth.label")

        # Plot results
        plot1 = fo.CategoricalHistogram("ground_truth.label", init_view=dataset)
        plot2 = fo.CategoricalHistogram(
            "ground_truth.label",
            order=lambda kv: -len(kv[0]),  # order by label length
            init_view=view1,
        )
        plot3 = fo.CategoricalHistogram("ground_truth.label", init_view=view2)
        plot = fo.ViewGrid([plot1, plot2, plot3])
        plot.show()

    Args:
        sample_collection: a :class:`fiftyone.core.collections.SampleCollection`
        k: the number of samples to select
        path: the categorical field against which to sample, e.g.,
            ``"ground_truth.label"``
        tag (None): an optional sample tag to use to encode the results
        exact (True): whether to tag exactly ``k`` samples (True) or sample so
            that the expected number of samples is ``k`` (False)
        seed (None): an optional random seed to use

    Returns:
        a :class:`fiftyone.core.view.DatasetView` containing the sample
    """
    counts = sample_collection.count_values(path)
    weights = np.array([1 / counts[l] for l in sample_collection.values(path)])

    return weighted_sample(
        sample_collection,
        k,
        weights,
        tag=tag,
        exact=exact,
        seed=seed,
    )



"""
Sphinx configuration file.

For a full list of available options, see:
https://www.sphinx-doc.org/en/master/usage/configuration.html

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import os
import re
import sys

sys.path.insert(0, os.path.abspath("."))

from custom_directives import (
    CustomButtonDirective,
    CustomCalloutItemDirective,
    CustomCardItemDirective,
    CustomImageLinkDirective,
)
from redirects import generate_redirects

import fiftyone.constants as foc


with open("../../setup.py") as f:
    setup_version = re.search(r'VERSION = "(.+?)"', f.read()).group(1)

if setup_version != foc.VERSION:
    raise RuntimeError(
        "FiftyOne version in setup.py (%r) does not match installed version "
        "(%r). If this is a dev install, reinstall with `pip install -e .` "
        "and try again." % (setup_version, foc.VERSION)
    )


# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#


# -- Project information -----------------------------------------------------

project = "FiftyOne"
copyright = foc.COPYRIGHT
author = foc.AUTHOR
release = foc.VERSION


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "nbsphinx",
    "sphinx_tabs.tabs",
    "sphinx_copybutton",
    "autodocsumm",
]

# Types of class members to generate documentation for.
autodoc_default_options = {
    "members": True,
    "inherited-members": True,
    "member-order": "bysource",
    "autosummary": True,
}
autodoc_inherit_docstrings = True
autoclass_content = "class"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of strings.
source_suffix = [".rst", ".md"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_includes"]

# A string of reStructuredText that will be included at the beginning of every
# source file that is read
rst_prolog = """
.. include:: /_includes/substitutions.rst
"""

# Disable nbshinx loading require.js - this breaks the pytorch theme's
# scrolling handling, and we don't appear to have any notebook content that
# requires it
nbsphinx_requirejs_path = ""

# Adds helpful external links to the built HTML
ref = "v%s" % foc.VERSION
nbsphinx_prolog = """

.. raw:: html

    <table class="fo-notebook-links" align="left">
        <td>
            <a target="_blank" href="https://colab.research.google.com/github/voxel51/fiftyone/blob/%s/docs/source/{{ env.doc2path(env.docname, base=None) }}">
                <img src="../_static/images/icons/colab-logo-256px.png"> &nbsp; Run in Google Colab
            </a>
        </td>
        <td>
            <a target="_blank" href="https://github.com/voxel51/fiftyone/blob/%s/docs/source/{{ env.doc2path(env.docname, base=None) }}">
                <img src="../_static/images/icons/github-logo-256px.png"> &nbsp; View source on GitHub
            </a>
        </td>
        <td>
            <a target="_blank" href="https://gitcdn.link/cdn/voxel51/fiftyone/%s/docs/source/{{ env.doc2path(env.docname, base=None) }}" download>
                <img src="../_static/images/icons/cloud-icon-256px.png"> &nbsp; Download notebook
            </a>
        </td>
    </table>

""" % (
    ref,
    ref,
    ref,
)

# Path to the redirects file, relative to `source/`
redirects_file = "redirects"

# -- Options for intersphinx extension ---------------------------------------

intersphinx_mapping = {
    # including `python` autolinks things like (None) which is not desirable
    # "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "plotly": ("https://plotly.com/python-api-reference/", None),
    "torch": ("https://pytorch.org/docs/stable/", None),
    "torchvision": ("https://pytorch.org/vision/stable/", None),
    "flash": ("https://lightning-flash.readthedocs.io/en/latest", None),
    "pymongo": ("https://pymongo.readthedocs.io/en/stable/", None),
    "mongoengine": ("https://docs.mongoengine.org/", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
    "pydicom": ("https://pydicom.github.io/pydicom/stable/", None),
    "rasterio": ("https://rasterio.readthedocs.io/en/latest/", None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pytorch_sphinx_theme"
html_theme_path = ["../theme"]
html_theme_options = {
    "pytorch_project": "docs",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = ["css/voxel51-website.css", "css/custom.css"]
html_js_files = ["js/voxel51-website.js", "js/custom.js"]

# Prevent RST source files from being included in output
html_copy_source = False

html_context = {
    "address_main_line1": "330 E Liberty St",
    "address_main_line2": "Ann Arbor, MI 48104",
    "phone_main": "+1 734-489-1134",
    "email_info": "info@voxel51.com",
    # Links - copied from website config
    "link_blog": "https://blog.voxel51.com/",
    "link_contactus": "mailto:solutions@voxel51.com?subject=[Voxel51]%20Contact%20us",
    "link_docs_fiftyone": "https://voxel51.com/docs/fiftyone/",
    "link_fiftyone": "https://voxel51.com/fiftyone/",
    "link_fiftyone_teams": "https://voxel51.com/fiftyone-teams/",
    "link_usecases": "https://voxel51.com/computer-vision-use-cases/",
    "link_fiftyone_tutorials": "https://voxel51.com/docs/fiftyone/tutorials/index.html",
    "link_fiftyone_examples": "https://github.com/voxel51/fiftyone-examples",
    "link_fiftyone_quickstart": "https://colab.research.google.com/github/voxel51/fiftyone-examples/blob/master/examples/quickstart.ipynb",
    "link_home": "https://voxel51.com/",
    "link_ourstory": "https://voxel51.com/ourstory/",
    "link_voxel51_jobs": "https://voxel51.com/jobs/",
    "link_press": "https://voxel51.com/press/",
    "link_privacypolicy": "https://voxel51.com/privacy/",
    "link_termsofservice": "https://voxel51.com/terms/",
    "link_voxel51_facebook": "https://www.facebook.com/voxel51/",
    "link_voxel51_github": "https://github.com/voxel51/",
    "link_voxel51_linkedin": "https://www.linkedin.com/company/voxel51/",
    "link_voxel51_slack": "https://join.slack.com/t/fiftyone-users/shared_invite/zt-s6936w7b-2R5eVPJoUw008wP7miJmPQ",
    "link_voxel51_twitter": "https://twitter.com/voxel51",
    "link_voxel51_medium": "https://medium.com/voxel51/",
    "banner_link": "https://voxel51.com/docs/fiftyone/integrations/coco.html",
}

# -- Custom app setup --------------------------------------------------------


def setup(app):
    # Generate page redirects
    app.add_config_value("redirects_file", "redirects", "env")
    app.connect("builder-inited", generate_redirects)

    # Custom directives
    app.add_directive("custombutton", CustomButtonDirective)
    app.add_directive("customcalloutitem", CustomCalloutItemDirective)
    app.add_directive("customcarditem", CustomCardItemDirective)
    app.add_directive("customimagelink", CustomImageLinkDirective)



"""
Package-wide constants.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from datetime import datetime
import os

from packaging.version import Version

try:
    from importlib.metadata import metadata  # Python 3.8
except ImportError:
    from importlib_metadata import metadata  # Python < 3.8


CLIENT_TYPE = "fiftyone"

FIFTYONE_DIR = os.path.dirname(os.path.abspath(__file__))
FIFTYONE_CONFIG_DIR = os.path.join(os.path.expanduser("~"), ".fiftyone")
FIFTYONE_CONFIG_PATH = os.path.join(FIFTYONE_CONFIG_DIR, "config.json")
FIFTYONE_ANNOTATION_CONFIG_PATH = os.path.join(
    FIFTYONE_CONFIG_DIR, "annotation_config.json"
)
FIFTYONE_APP_CONFIG_PATH = os.path.join(FIFTYONE_CONFIG_DIR, "app_config.json")
BASE_DIR = os.path.dirname(FIFTYONE_DIR)
TEAMS_PATH = os.path.join(FIFTYONE_CONFIG_DIR, "var", "teams.json")
RESOURCES_DIR = os.path.join(FIFTYONE_DIR, "resources")

#
# The compatible versions for this client
#
# RULES: Datasets from all compatible versions must be...
#   - Loadable by this client without error
#   - Editable by this client without causing any database edits that would
#     break the original client's ability to work with the dataset
#
# This setting may be ``None`` if this client has no compatibility with other
# versions
#
COMPATIBLE_VERSIONS = ">=0.16.6,<0.19"

# Package metadata
_META = metadata("fiftyone")
NAME = _META["name"]
VERSION = _META["version"]
DESCRIPTION = _META["summary"]
AUTHOR = _META["author"]
AUTHOR_EMAIL = _META["author-email"]
URL = _META["home-page"]
LICENSE = _META["license"]
VERSION_LONG = "FiftyOne v%s, %s" % (VERSION, AUTHOR)
COPYRIGHT = "2017-%d, %s" % (datetime.now().year, AUTHOR)

DEV_INSTALL = os.path.isdir(
    os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", ".git")
    )
)
RC_INSTALL = "rc" in VERSION

# App configuration
DEFAULT_APP_COLOR_POOL = [
    "#ee0000",
    "#ee6600",
    "#993300",
    "#996633",
    "#999900",
    "#009900",
    "#003300",
    "#009999",
    "#000099",
    "#0066ff",
    "#6600ff",
    "#cc33cc",
    "#777799",
]

# MongoDB setup
try:
    from fiftyone.db import FIFTYONE_DB_BIN_DIR
except ImportError:
    # development installation
    FIFTYONE_DB_BIN_DIR = os.path.join(FIFTYONE_CONFIG_DIR, "bin")

DEFAULT_DB_DIR = os.path.join(FIFTYONE_CONFIG_DIR, "var", "lib", "mongo")
MIGRATIONS_PATH = os.path.join(FIFTYONE_CONFIG_DIR, "migrations")
MIGRATIONS_HEAD_PATH = os.path.join(MIGRATIONS_PATH, "head.json")
MIGRATIONS_REVISIONS_DIR = os.path.join(
    FIFTYONE_DIR, "migrations", "revisions"
)

MIN_MONGODB_VERSION = Version("4.4")
DATABASE_APPNAME = "fiftyone"

# Server setup
SERVER_DIR = os.path.join(FIFTYONE_DIR, "server")

# App setup
try:
    from fiftyone.desktop import FIFTYONE_DESKTOP_APP_DIR
except ImportError:
    FIFTYONE_DESKTOP_APP_DIR = os.path.normpath(
        os.path.join(FIFTYONE_DIR, "../app")
    )

# Analytics
UA_DEV = "UA-141773487-10"
UA_PROD = "UA-141773487-9"
UA_ID = UA_DEV if DEV_INSTALL or RC_INSTALL else UA_PROD



"""
FiftyOne Server view

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import fiftyone.core.dataset as fod
from fiftyone.core.expressions import ViewField as F, VALUE
import fiftyone.core.fields as fof
import fiftyone.core.labels as fol
import fiftyone.core.media as fom
import fiftyone.core.stages as fosg
import fiftyone.core.utils as fou
import fiftyone.core.view as fov

from fiftyone.server.utils import iter_label_fields


_LABEL_TAGS = "_label_tags"


def get_view(
    dataset_name,
    stages=None,
    filters=None,
    count_label_tags=False,
    only_matches=True,
    extended_stages=None,
    sample_filter=None,
    sort=False,
):
    """Get the view from request paramters

    Args:
        dataset_names: the dataset name
        stages (None): an optional list of serialized
            :class:`fiftyone.core.stages.ViewStage` instances
        filters (None): an optional ``dict`` of App defined filters
        extended_stages (None): extended view stages
        count_label_tags (False): whether to set the hidden ``_label_tags``
            field with counts of tags with respect to all label fields
        only_matches (True): whether to filter unmatches samples when filtering
            labels
        sample_filter (None): an optional
            :class:`fiftyone.server.filters.SampleFilter`
        sort (False): whether to include sort extended stages

    Returns:
        a :class:`fiftyone.core.view.DatasetView`
    """
    view = fod.load_dataset(dataset_name).view()
    view.reload()

    if stages:
        view = fov.DatasetView._build(view, stages)

    if sample_filter is not None:
        if sample_filter.group:
            if sample_filter.group.id:
                view = fov.make_optimized_select_view(
                    view, sample_filter.group.id, groups=True
                )
            if sample_filter.group.slice:
                view.group_slice = sample_filter.group.slice

        elif sample_filter.id:
            view = fov.make_optimized_select_view(view, sample_filter.id)

    if filters or extended_stages or count_label_tags:
        view = get_extended_view(
            view,
            filters,
            count_label_tags=count_label_tags,
            only_matches=only_matches,
            extended_stages=extended_stages,
            sort=sort,
        )

    return view


def get_extended_view(
    view,
    filters=None,
    count_label_tags=False,
    only_matches=True,
    extended_stages=None,
    sort=False,
):
    """Create an extended view with the provided filters.

    Args:
        view: a :class:`fiftyone.core.collections.SampleCollection`
        filters: an optional ``dict`` of App defined filters
        count_label_tags (False): whether to set the hidden ``_label_tags``
            field with counts of tags with respect to all label fields
        only_matches (True): whether to filter unmatches samples when filtering
            labels
        extended_stages (None): extended view stages
        sort (False): wheter to include sort extended stages
    """
    cleanup_fields = set()
    filtered_labels = set()

    label_tags = None
    if filters is not None and len(filters):
        if "tags" in filters:
            tags = filters.get("tags")
            if "label" in tags:
                label_tags = tags["label"]

            if not count_label_tags and label_tags:
                view = view.select_labels(tags=label_tags)

            if "sample" in tags:
                view = view.match_tags(tags=tags["sample"])

        stages, cleanup_fields, filtered_labels = _make_filter_stages(
            view,
            filters,
            label_tags=label_tags,
            hide_result=count_label_tags,
            only_matches=only_matches,
        )

        for stage in stages:
            view = view.add_stage(stage)

    if extended_stages:
        view = extend_view(view, extended_stages, sort)

    if count_label_tags:
        view = _add_labels_tags_counts(view, filtered_labels, label_tags)
        if cleanup_fields:
            view = view.mongo([{"$unset": field} for field in cleanup_fields])

    return view


def extend_view(view, extended_stages, sort):
    for cls, d in extended_stages.items():
        kwargs = [[k, v] for k, v in d.items()]
        stage = fosg.ViewStage._from_dict({"_cls": cls, "kwargs": kwargs})
        if sort or not isinstance(stage, (fosg.SortBySimilarity, fosg.SortBy)):
            view = view.add_stage(stage)

    return view


def _add_labels_tags_counts(view, filtered_fields, label_tags):
    view = view.set_field(_LABEL_TAGS, [], _allow_missing=True)

    for path, field in iter_label_fields(view):
        if not issubclass(
            field.document_type, (fol._HasID, fol._HasLabelList)
        ):
            continue

        path = _get_filtered_path(view, path, filtered_fields, label_tags)
        if issubclass(field.document_type, fol._HasLabelList):
            if path.startswith(view._FRAMES_PREFIX):
                add_tags = _add_frame_labels_tags
            else:
                add_tags = _add_labels_tags
        else:
            if path.startswith(view._FRAMES_PREFIX):
                add_tags = _add_frame_label_tags
            else:
                add_tags = _add_label_tags

        view = add_tags(path, field, view)

    view = _count_list_items(_LABEL_TAGS, view)

    return view


def _make_expression(field, path, args):
    if not path:
        return _make_scalar_expression(F(), args, field)

    keys = path.split(".")
    rest = ".".join(keys[1:])
    field = field.get_field_schema()[keys[0]]
    if isinstance(field, fof.ListField) and not isinstance(
        field, fof.FrameSupportField
    ):
        new_field = field.field
        expr = (
            lambda subexpr: F(field.db_field or field.name)
            .filter(subexpr)
            .length()
            > 0
        )
    else:
        new_field = field
        expr = lambda subexpr: F(field.db_field or field.name).apply(subexpr)

    subexpr = _make_expression(new_field, rest, args)
    if subexpr is not None:
        return expr(subexpr)

    return None


def _make_filter_stages(
    view, filters, label_tags=None, hide_result=False, only_matches=True
):
    field_schema = view.get_field_schema()
    if view.media_type != fom.IMAGE:
        frame_field_schema = view.get_frame_field_schema()
    else:
        frame_field_schema = None

    tag_expr = (F("tags") != None).if_else(
        F("tags").contains(label_tags), None
    )
    cache = {}

    stages = []
    cleanup = set()
    filtered_labels = set()
    for path in sorted(filters):
        args = filters[path]
        if path == "tags" or path.startswith("_"):
            continue

        frames = path.startswith(view._FRAMES_PREFIX)
        keys = path.split(".")
        if frames:
            field = frame_field_schema[keys[1]]
            keys = keys[2:]
            prefix = "frames."
        else:
            field = field_schema[keys[0]]
            keys = keys[1:]
            prefix = ""

        if _is_label(field):
            keypoints = issubclass(
                field.document_type, (fol.Keypoint, fol.Keypoints)
            )

            if keypoints:
                if path.endswith(".id") or path.endswith(".label"):
                    keypoints = False

            parent = field
            field = view.get_field(path)
            key = field.db_field if field.db_field else field.name
            view_field = F(key)
            expr = (
                _make_keypoint_kwargs(
                    args,
                    view,
                    path if path.endswith(".points") else None,
                )
                if keypoints
                else _make_scalar_expression(view_field, args, field)
            )
            if expr is not None:
                if hide_result:
                    new_field = "__%s" % path.split(".")[1 if frames else 0]
                    if frames:
                        new_field = "%s%s" % (
                            view._FRAMES_PREFIX,
                            new_field,
                        )
                else:
                    new_field = None

                if keypoints:
                    stage = fosg.FilterKeypoints(
                        prefix + parent.name,
                        _new_field=new_field,
                        **expr,
                    )
                else:
                    stage = fosg.FilterLabels(
                        cache.get(prefix + parent.name, prefix + parent.name),
                        expr,
                        only_matches=only_matches,
                        _new_field=new_field,
                    )

                stages.append(stage)
                filtered_labels.add(path)
                if new_field:
                    cache[prefix + parent.name] = new_field
                    cleanup.add(new_field)
        else:
            expr = _make_expression(view, path, args)
            if expr is not None:
                stages.append(fosg.Match(expr))

    if label_tags is not None and hide_result:
        for path, _ in iter_label_fields(view):
            if hide_result:
                new_field = _get_filtered_path(
                    view, path, filtered_labels, label_tags
                )
            else:
                new_field = None

            stages.append(
                fosg.FilterLabels(
                    cache.get(path, path),
                    tag_expr,
                    only_matches=False,
                    _new_field=new_field,
                )
            )
            if new_field:
                cache[path] = new_field
                cleanup.add(new_field)

        match_exprs = []
        for path, _ in iter_label_fields(view):
            match_exprs.append(
                fosg._get_label_field_only_matches_expr(
                    view,
                    cache.get(path, path),
                )
            )

        stages.append(fosg.Match(F.any(match_exprs)))

    return stages, cleanup, filtered_labels


def _is_support(field):
    if isinstance(field, fof.FrameSupportField):
        return True

    if isinstance(field, fof.EmbeddedDocumentField):
        if field.document_type in (
            fol.TemporalDetection,
            fol.TemporalDetections,
        ):
            return True

    return False


def _is_datetime(field):
    return isinstance(field, (fof.DateField, fof.DateTimeField))


def _is_label(field):
    return isinstance(field, fof.EmbeddedDocumentField) and issubclass(
        field.document_type, fol.Label
    )


def _make_scalar_expression(f, args, field):
    expr = None
    if isinstance(field, fof.BooleanField):
        true, false = args["true"], args["false"]
        if true and false:
            expr = f.is_in([True, False])

        if not true and false:
            expr = f == False

        if true and not false:
            expr = f == True

        if not true and not false:
            expr = (f != True) & (f != False)

    elif _is_support(field):
        mn, mx = args["range"]
        expr = (f[0] >= mn) & (f[1] <= mx)
    elif _is_datetime(field):
        mn, mx = args["range"]
        p = fou.timestamp_to_datetime
        expr = (f >= p(mn)) & (f <= p(mx))
    elif isinstance(field, (fof.FloatField, fof.IntField)):
        mn, mx = args["range"]
        expr = (f >= mn) & (f <= mx)
    else:
        values = args["values"]
        if not values:
            return None

        if isinstance(field, fof.ObjectIdField):
            f = f.to_string()

        none = any(map(lambda v: v is None, values))
        values = filter(lambda v: v is not None, values)
        expr = f.is_in(values)
        exclude = args["exclude"]

        if exclude:
            # pylint: disable=invalid-unary-operand-type
            expr = ~expr

        if none:
            if exclude:
                expr &= f.exists()
            else:
                expr |= ~(f.exists())

        return expr

    return _apply_others(expr, f, args)


def _make_keypoint_kwargs(args, view, points):
    if points:
        ske = view._dataset.default_skeleton
        name = points.split(".")[0]

        if name in view._dataset.skeletons:
            ske = view._dataset.skeletons[name]

        values = args.get("values", [])
        if args["exclude"]:
            values = set(ske.labels).difference(values)

        return {"labels": values}

    f = F("confidence")
    mn, mx = args["range"]
    return {"filter": (f >= mn) & (f <= mx)}


def _apply_others(expr, f, args):
    nonfinites = {
        "nan": float("nan"),
        "ninf": -float("inf"),
        "inf": float("inf"),
    }
    include = []
    for k, v in nonfinites.items():
        if k in args and args[k]:
            include.append(v)

    if expr is None:
        expr = f.is_in(include)
    else:
        expr |= f.is_in(include)

    if "none" in args:
        expr = _apply_none(expr, f, args["none"])

    if "exclude" in args and args["exclude"]:
        # pylint: disable=invalid-unary-operand-type
        expr = ~expr

    return expr


def _apply_none(expr, f, none):
    if not none:
        if expr is not None:
            expr &= f.exists()
        else:
            expr = f.exists()
    elif expr is not None:
        expr |= ~(f.exists())

    return expr


def _get_filtered_path(view, path, filtered_fields, label_tags):
    if path not in filtered_fields and not label_tags:
        return path

    if path.startswith(view._FRAMES_PREFIX):
        return "%s__%s" % (view._FRAMES_PREFIX, path.split(".")[1])

    return "__%s" % path


def _add_frame_labels_tags(path, field, view):
    frames, path = path.split(".")
    items = "%s.%s" % (path, field.document_type._LABEL_LIST_FIELD)
    view = view.set_field(
        _LABEL_TAGS,
        F(_LABEL_TAGS).extend(
            F(frames).reduce(
                VALUE.extend(F(items).reduce(VALUE.extend(F("tags")), [])), []
            )
        ),
        _allow_missing=True,
    )
    return view


def _add_frame_label_tags(path, field, view):
    frames, path = path.split(".")
    tags = "%s.tags" % path
    view = view.set_field(
        _LABEL_TAGS,
        F(_LABEL_TAGS).extend(
            F(frames).reduce(
                VALUE.extend((F(tags) != None).if_else(F(tags), [])), []
            )
        ),
        _allow_missing=True,
    )
    return view


def _add_labels_tags(path, field, view):
    items = "%s.%s" % (path, field.document_type._LABEL_LIST_FIELD)
    view = view.set_field(
        _LABEL_TAGS,
        F(_LABEL_TAGS).extend(F(items).reduce(VALUE.extend(F("tags")), [])),
        _allow_missing=True,
    )
    return view


def _add_label_tags(path, field, view):
    tags = "%s.tags" % path
    return view.set_field(
        _LABEL_TAGS,
        F(_LABEL_TAGS).extend((F(tags) != None).if_else(F(tags), [])),
        _allow_missing=True,
    )


def _count_list_items(path, view):
    function = (
        "function(items) {"
        "let counts = {};"
        "items && items.forEach((i) => {"
        "counts[i] = 1 + (counts[i] || 0);"
        "});"
        "return counts;"
        "}"
    )

    return view.set_field(
        path, F(path)._function(function), _allow_missing=True
    )



"""
FiftyOne Server /frames route

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request

from fiftyone.core.expressions import ViewField as F
import fiftyone.core.json as foj
import fiftyone.core.odm as foo
import fiftyone.core.view as fov

from fiftyone.server.decorators import route
import fiftyone.server.view as fosv


class Frames(HTTPEndpoint):
    @route
    async def post(self, request: Request, data: dict):
        start_frame = int(data.get("frameNumber", 1))
        frame_count = int(data.get("frameCount", 1))
        num_frames = int(data.get("numFrames"))
        extended = data.get("extended", None)
        dataset = data.get("dataset")
        stages = data.get("view")
        sample_id = data.get("sampleId")

        view = fosv.get_view(dataset, stages=stages, extended_stages=extended)
        view = fov.make_optimized_select_view(view, sample_id)

        end_frame = min(num_frames + start_frame, frame_count)
        frames = await foo.aggregate(
            foo.get_async_db_conn()[view._dataset._sample_collection_name],
            view._pipeline(frames_only=True, support=[start_frame, end_frame]),
        ).to_list(end_frame - start_frame + 1)

        return {
            "frames": foj.stringify(frames),
            "range": [start_frame, end_frame],
        }



"""
Dataset ingestors.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

import eta.core.image as etai
import eta.core.utils as etau

import fiftyone as fo
import fiftyone.core.utils as fou

from .importers import (
    UnlabeledImageDatasetImporter,
    LabeledImageDatasetImporter,
    UnlabeledVideoDatasetImporter,
    LabeledVideoDatasetImporter,
)


logger = logging.getLogger(__name__)


class ImageIngestor(object):
    """Mixin for :class:`fiftyone.utils.data.importers.DatasetImporter`
    instances that ingest images into the provided ``dataset_dir`` during
    import.

    Args:
        dataset_dir: the directory where input images will be ingested into
        image_format (None): the image format to use when writing in-memory
            images to disk. By default, ``fiftyone.config.default_image_ext``
            is used
    """

    def __init__(self, dataset_dir, image_format=None):
        if image_format is None:
            image_format = fo.config.default_image_ext

        self.dataset_dir = dataset_dir
        self.image_format = image_format

        self._filename_maker = None

    def _ingest_image(self, sample_parser):
        if sample_parser.has_image_path:
            try:
                return self._ingest_image_from_path(sample_parser)
            except:
                # Allow for SampleParsers that declare `has_image_path == True`
                # but cannot generate paths at runtime, e.g., because they
                # support inputs of the form `image_or_path` and an image, not
                # a path, was provided
                pass

        return self._ingest_in_memory_image(sample_parser)

    def _ingest_image_from_path(self, sample_parser):
        image_path = sample_parser.get_image_path()
        output_image_path = self._filename_maker.get_output_path(image_path)
        etau.copy_file(image_path, output_image_path)
        return output_image_path

    def _ingest_in_memory_image(self, sample_parser):
        img = sample_parser.get_image()
        image_path = self._filename_maker.get_output_path()
        etai.write(img, image_path)
        return image_path

    def _setup(self):
        self._filename_maker = fou.UniqueFilenameMaker(
            output_dir=self.dataset_dir, default_ext=self.image_format
        )


class UnlabeledImageDatasetIngestor(
    UnlabeledImageDatasetImporter, ImageIngestor
):
    """Dataset importer that ingests unlabeled images into the provided
    ``dataset_dir`` during import.

    The source images are parsed from the provided ``samples`` using the
    provided :class:`fiftyone.utils.data.parsers.UnlabeledImageSampleParser`.

    If an image path is available via
    :func:`fiftyone.utils.data.parsers.UnlabeledImageSampleParser.get_image_path`,
    then the image is directly copied from its source location into
    ``dataset_dir``. In this case, the original filename is maintained, unless
    a name conflict would occur, in which case an index of the form
    ``"-%d" % count`` is appended to the base filename.

    If no image path is available, the image is read in-memory via
    :func:`fiftyone.utils.data.parsers.UnlabeledImageSampleParser.get_image`
    and written to ``dataset_dir`` in the following format::

        <dataset_dir>/<image_count><image_format>

    where ``image_count`` is the number of files in ``dataset_dir``.

    Args:
        dataset_dir: the directory where input images will be ingested into
        samples: an iterable of samples that can be parsed by ``sample_parser``
        sample_parser: an
            :class:`fiftyone.utils.data.parsers.UnlabeledImageSampleParser` to
            use to parse the samples
        image_format (None): the image format to use when writing in-memory
            images to disk. By default, ``fiftyone.config.default_image_ext``
            is used
        max_samples (None): a maximum number of samples to import. By default,
            all samples are imported
    """

    def __init__(
        self,
        dataset_dir,
        samples,
        sample_parser,
        image_format=None,
        max_samples=None,
    ):
        UnlabeledImageDatasetImporter.__init__(
            self, dataset_dir=dataset_dir, max_samples=max_samples
        )
        ImageIngestor.__init__(self, dataset_dir, image_format=image_format)

        self.samples = samples
        self.sample_parser = sample_parser

        self._iter_samples = None
        self._num_samples = None
        self._num_imported = None

    def __iter__(self):
        self._num_imported = 0
        self._iter_samples = iter(self.samples)
        return self

    def __len__(self):
        if self._num_samples is not None:
            return self._num_samples

        return len(self.samples)

    def __next__(self):
        if (
            self.max_samples is not None
            and self._num_imported >= self.max_samples
        ):
            raise StopIteration

        sample = next(self._iter_samples)

        self.sample_parser.with_sample(sample)

        image_path = self._ingest_image(self.sample_parser)

        if self.has_image_metadata:
            image_metadata = self.sample_parser.get_image_metadata()
        else:
            image_metadata = None

        self._num_imported += 1

        return image_path, image_metadata

    @property
    def has_dataset_info(self):
        return False

    @property
    def has_image_metadata(self):
        return self.sample_parser.has_image_metadata

    def setup(self):
        self._setup()

        try:
            self._num_samples = len(self.samples)
            if self.max_samples is not None:
                self._num_samples = min(self._num_samples, self.max_samples)
        except:
            pass


class LabeledImageDatasetIngestor(LabeledImageDatasetImporter, ImageIngestor):
    """Dataset importer that ingests labeled images into the provided
    ``dataset_dir`` during import.

    The source images and labels are parsed from the provided ``samples`` using
    the provided :class:`fiftyone.utils.data.parsers.LabeledImageSampleParser`.

    If an image path is available via
    :func:`fiftyone.utils.data.parsers.LabeledImageSampleParser.get_image_path`,
    then the image is directly copied from its source location into
    ``dataset_dir``. In this case, the original filename is maintained, unless
    a name conflict would occur, in which case an index of the form
    ``"-%d" % count`` is appended to the base filename.

    If no image path is available, the image is read in-memory via
    :func:`fiftyone.utils.data.parsers.LabeledImageSampleParser.get_image` and
    written to ``dataset_dir`` in the following format::

        <dataset_dir>/<image_count><image_format>

    where ``image_count`` is the number of files in ``dataset_dir``.

    Args:
        dataset_dir: the directory where input images will be ingested into
        samples: an iterable of samples that can be parsed by ``sample_parser``
        sample_parser: an
            :class:`fiftyone.utils.data.parsers.LabeledImageSampleParser` to
            use to parse the samples
        image_format (None): the image format to use when writing in-memory
            images to disk. By default, ``fiftyone.config.default_image_ext``
            is used
        max_samples (None): a maximum number of samples to import. By default,
            all samples are imported
    """

    def __init__(
        self,
        dataset_dir,
        samples,
        sample_parser,
        image_format=None,
        max_samples=None,
    ):
        LabeledImageDatasetImporter.__init__(
            self, dataset_dir=dataset_dir, max_samples=max_samples
        )
        ImageIngestor.__init__(self, dataset_dir, image_format=image_format)

        self.samples = samples
        self.sample_parser = sample_parser

        self._iter_samples = None
        self._num_samples = None
        self._num_imported = None

    def __iter__(self):
        self._num_imported = 0
        self._iter_samples = iter(self.samples)
        return self

    def __len__(self):
        if self._num_samples is not None:
            return self._num_samples

        return len(self.samples)

    def __next__(self):
        if (
            self.max_samples is not None
            and self._num_imported >= self.max_samples
        ):
            raise StopIteration

        sample = next(self._iter_samples)

        self.sample_parser.with_sample(sample)

        image_path = self._ingest_image(self.sample_parser)

        if self.has_image_metadata:
            image_metadata = self.sample_parser.get_image_metadata()
        else:
            image_metadata = None

        label = self.sample_parser.get_label()

        self._num_imported += 1

        return image_path, image_metadata, label

    @property
    def has_dataset_info(self):
        return False

    @property
    def has_image_metadata(self):
        return self.sample_parser.has_image_metadata

    @property
    def label_cls(self):
        return self.sample_parser.label_cls

    def setup(self):
        self._setup()

        try:
            self._num_samples = len(self.samples)
            if self.max_samples is not None:
                self._num_samples = min(self._num_samples, self.max_samples)
        except:
            pass


class VideoIngestor(object):
    """Mixin for :class:`fiftyone.utils.data.importers.DatasetImporter`
    instances that ingest videos into the provided ``dataset_dir`` during
    import.

    Args:
        dataset_dir: the directory where input videos will be ingested into
    """

    def __init__(self, dataset_dir):
        self.dataset_dir = dataset_dir

        self._filename_maker = None

    def _ingest_video(self, sample_parser):
        video_path = sample_parser.get_video_path()
        output_video_path = self._filename_maker.get_output_path(video_path)
        etau.copy_file(video_path, output_video_path)
        return output_video_path

    def _setup(self):
        self._filename_maker = fou.UniqueFilenameMaker(
            output_dir=self.dataset_dir
        )


class UnlabeledVideoDatasetIngestor(
    UnlabeledVideoDatasetImporter, VideoIngestor
):
    """Dataset importer that ingests unlabeled videos into the provided
    ``dataset_dir`` during import.

    The source videos are parsed from the provided ``samples`` using the
    provided :class:`fiftyone.utils.data.parsers.UnlabeledVideoSampleParser`.

    The source videos are directly copied from their source locations into
    ``dataset_dir``, maintaining the original filenames, unless a name conflict
    would occur, in which case an index of the form ``"-%d" % count`` is
    appended to the base filename.

    Args:
        dataset_dir: the directory where input videos will be ingested into
        samples: an iterable of samples that can be parsed by ``sample_parser``
        sample_parser: an
            :class:`fiftyone.utils.data.parsers.UnlabeledVideoSampleParser` to
            use to parse the samples
        max_samples (None): a maximum number of samples to import. By default,
            all samples are imported
    """

    def __init__(self, dataset_dir, samples, sample_parser, max_samples=None):
        UnlabeledVideoDatasetImporter.__init__(
            self, dataset_dir=dataset_dir, max_samples=max_samples
        )
        VideoIngestor.__init__(self, dataset_dir)

        self.samples = samples
        self.sample_parser = sample_parser

        self._iter_samples = None
        self._num_samples = None
        self._num_imported = None

    def __iter__(self):
        self._num_imported = 0
        self._iter_samples = iter(self.samples)
        return self

    def __len__(self):
        if self._num_samples is not None:
            return self._num_samples

        return len(self.samples)

    def __next__(self):
        if (
            self.max_samples is not None
            and self._num_imported >= self.max_samples
        ):
            raise StopIteration

        sample = next(self._iter_samples)

        self.sample_parser.with_sample(sample)

        video_path = self._ingest_video(self.sample_parser)

        if self.has_video_metadata:
            video_metadata = self.sample_parser.get_video_metadata()
        else:
            video_metadata = None

        self._num_imported += 1

        return video_path, video_metadata

    @property
    def has_dataset_info(self):
        return False

    @property
    def has_video_metadata(self):
        return self.sample_parser.has_video_metadata

    def setup(self):
        self._setup()

        try:
            self._num_samples = len(self.samples)
            if self.max_samples is not None:
                self._num_samples = min(self._num_samples, self.max_samples)
        except:
            pass


class LabeledVideoDatasetIngestor(LabeledVideoDatasetImporter, VideoIngestor):
    """Dataset importer that ingests labeled videos into the provided
    ``dataset_dir`` during import.

    The source videos and labels are parsed from the provided ``samples`` using
    the provided :class:`fiftyone.utils.data.parsers.LabeledVideoSampleParser`.

    The source videos are directly copied from their source locations into
    ``dataset_dir``, maintaining the original filenames, unless a name conflict
    would occur, in which case an index of the form ``"-%d" % count`` is
    appended to the base filename.

    Args:
        dataset_dir: the directory where input videos will be ingested into
        samples: an iterable of samples that can be parsed by ``sample_parser``
        sample_parser: an
            :class:`fiftyone.utils.data.parsers.LabeledVideoSampleParser` to
            use to parse the samples
        max_samples (None): a maximum number of samples to import. By default,
            all samples are imported
    """

    def __init__(
        self,
        dataset_dir,
        samples,
        sample_parser,
        max_samples=None,
    ):
        LabeledVideoDatasetImporter.__init__(
            self, dataset_dir=dataset_dir, max_samples=max_samples
        )
        VideoIngestor.__init__(self, dataset_dir)

        self.samples = samples
        self.sample_parser = sample_parser

        self._iter_samples = None
        self._num_samples = None
        self._num_imported = None

    def __iter__(self):
        self._num_imported = 0
        self._iter_samples = iter(self.samples)
        return self

    def __len__(self):
        if self._num_samples is not None:
            return self._num_samples

        return len(self.samples)

    def __next__(self):
        if (
            self.max_samples is not None
            and self._num_imported >= self.max_samples
        ):
            raise StopIteration

        sample = next(self._iter_samples)

        self.sample_parser.with_sample(sample)

        video_path = self._ingest_video(self.sample_parser)

        if self.has_video_metadata:
            video_metadata = self.sample_parser.has_video_metadata()
        else:
            video_metadata = None

        label = self.sample_parser.get_label()
        frames = self.sample_parser.get_frame_labels()

        self._num_imported += 1

        return video_path, video_metadata, label, frames

    @property
    def has_dataset_info(self):
        return False

    @property
    def has_video_metadata(self):
        return self.sample_parser.has_video_metadata

    @property
    def label_cls(self):
        return self.sample_parser.label_cls

    @property
    def frame_labels_cls(self):
        return self.sample_parser.frame_labels_cls

    def setup(self):
        self._setup()

        try:
            self._num_samples = len(self.samples)
            if self.max_samples is not None:
                self._num_samples = min(self._num_samples, self.max_samples)
        except:
            pass



"""
FiftyOne v0.13.0 revision.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""


def up(db, dataset_name):
    match_d = {"name": dataset_name}
    dataset_dict = db.datasets.find_one(match_d)

    if "annotation_runs" not in dataset_dict:
        dataset_dict["annotation_runs"] = {}

    brain_methods = dataset_dict.get("brain_methods", {})
    for run_doc in brain_methods.values():
        config = run_doc.get("config", {})
        config_cls = config.get("cls", "")
        if config_cls == "fiftyone.brain.similarity.SimilarityConfig":
            config["model"] = None
            config["metric"] = "euclidean"

    db.datasets.replace_one(match_d, dataset_dict)


def down(db, dataset_name):
    match_d = {"name": dataset_name}
    dataset_dict = db.datasets.find_one(match_d)

    dataset_dict.pop("annotation_runs", None)

    brain_methods = dataset_dict.get("brain_methods", {})
    for run_doc in brain_methods.values():
        config = run_doc.get("config", {})
        config_cls = config.get("cls", "")
        if config_cls == "fiftyone.brain.similarity.SimilarityConfig":
            config.pop("model", None)
            config.pop("metric", None)

    db.datasets.replace_one(match_d, dataset_dict)



"""
Evaluation runs framework.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from fiftyone.core.runs import Run, RunInfo, RunConfig, RunResults


class EvaluationInfo(RunInfo):
    """Information about an evaluation that has been run on a dataset.

    Args:
        key: the evaluation key
        timestamp (None): the UTC ``datetime`` when the evaluation was run
        config (None): the :class:`EvaluationMethodConfig` for the evaluation
    """

    @classmethod
    def config_cls(cls):
        return EvaluationMethodConfig


class EvaluationMethodConfig(RunConfig):
    """Base class for configuring :class:`EvaluationMethod` instances.

    Args:
        **kwargs: any leftover keyword arguments after subclasses have done
            their parsing
    """

    pass


class EvaluationMethod(Run):
    """Base class for evaluation methods.

    Subclasses will typically declare an interface method that handles
    performing evaluation on an image, video, or entire collection.

    Args:
        config: an :class:`EvaluationMethodConfig`
    """

    @classmethod
    def run_info_cls(cls):
        return EvaluationInfo

    @classmethod
    def _runs_field(cls):
        return "evaluations"

    @classmethod
    def _run_str(cls):
        return "evaluation"

    @classmethod
    def _results_cache_field(cls):
        return "_evaluation_cache"


class EvaluationResults(RunResults):
    """Base class for evaluation results."""

    pass



"""
FiftyOne v0.7.2 revision.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import pymongo as pm


def up(db, dataset_name):
    match_d = {"name": dataset_name}
    dataset_dict = db.datasets.find_one(match_d)
    if dataset_dict["media_type"] != "video":
        return

    fields = []
    for field in dataset_dict["sample_fields"]:
        if field["name"] == "frames":
            continue

        fields.append(field)

    dataset_dict["sample_fields"] = fields
    db.datasets.replace_one(match_d, dataset_dict)
    sample_coll = db[dataset_dict["sample_collection_name"]]
    sample_coll.update_many({}, {"$unset": {"frames": ""}})


def down(db, dataset_name):
    match_d = {"name": dataset_name}
    dataset_dict = db.datasets.find_one(match_d)
    if dataset_dict["media_type"] != "video":
        return

    frames = {
        "name": "frames",
        "ftype": "fiftyone.core.fields.EmbeddedDocumentField",
        "subfield": None,
        "embedded_doc_type": "fiftyone.core.labels._Frames",
    }
    dataset_dict["sample_fields"].append(frames)
    db.datasets.replace_one(match_d, dataset_dict)
    frame_coll = db["frames." + dataset_dict["sample_collection_name"]]
    sample_coll = db[dataset_dict["sample_collection_name"]]

    first_frames = frame_coll.find({"frame_number": 1})
    writes = []
    for f in first_frames:
        frame_d = {"first_frame": f}
        writes.append(
            pm.UpdateOne(
                {"_id": f["_sample_id"]}, {"$set": {"frames": frame_d}}
            )
        )

    sample_coll.bulk_write(writes)

    counts = frame_coll.aggregate(
        [{"$group": {"_id": "$_sample_id", "count": {"$sum": 1}}}]
    )
    writes = []
    for count in counts:
        writes.append(
            pm.UpdateOne(
                {"_id": count["_id"]},
                {"$set": {"frames.frame_count": count["count"]}},
            )
        )

    sample_coll.bulk_write(writes)



"""
Utilities for working with `Amazon Web Services <https://aws.amazon.com>`.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging
import multiprocessing
import multiprocessing.dummy
import os
from urllib.parse import urlparse

import boto3
import botocore

import eta.core.utils as etau

import fiftyone.core.utils as fou


logger = logging.getLogger(__name__)


def download_public_s3_files(
    urls, download_dir=None, num_workers=None, overwrite=True
):
    """Download files from a public AWS S3 bucket using unsigned URLs.

    The `url` argument either accepts:

        -   A list of paths to objects in the s3 bucket::

                urls = ["s3://bucket_name/dir1/file1.ext", ...]

            When `urls` is a list, then the `download_dir` argument is required
            and all objects will be downloaded into that directory

        -   A dictionary mapping the paths of objects to files on disk to store
            each object::

                urls = {
                    "s3://bucket_name/dir1/file1.ext": "/path/to/local/file1.ext",
                    ...
                }

    Args:
        urls: either a list of URLs to objects in an s3 bucket, or a dict
            mapping these URLs to locations on disk. If `urls` is a list, then
            the `download_dir` argument is required
        download_dir (None): the directory to store all downloaded objects.
            This is only used if `urls` is a list
        num_workers (None): the number of processes to use when downloading
            files. By default, ``multiprocessing.cpu_count()`` is used
        overwrite (True): whether to overwrite existing files
    """
    if not isinstance(urls, dict):
        if download_dir is None:
            raise ValueError(
                "`download_dir` is required when `urls` is a list"
            )

        urls = {url: None for url in urls}

    if download_dir:
        etau.ensure_dir(download_dir)

    if num_workers is None:
        num_workers = multiprocessing.cpu_count()

    s3_client = boto3.client(
        "s3",
        config=botocore.config.Config(
            signature_version=botocore.UNSIGNED,
            max_pool_connections=max(10, num_workers),
        ),
    )

    inputs = _build_inputs(
        urls, s3_client, download_dir=download_dir, overwrite=overwrite
    )

    if not inputs:
        return

    if num_workers <= 1:
        _single_thread_download(inputs)
    else:
        _multi_thread_download(inputs, num_workers)


def _build_inputs(urls, s3_client, download_dir=None, overwrite=True):
    inputs = []
    for url, filepath in urls.items():
        bucket_name, object_path = _parse_url(url)
        if filepath is None:
            filepath = os.path.join(download_dir, object_path)

        if not os.path.isfile(filepath):
            inputs.append((bucket_name, object_path, filepath, s3_client))
        else:
            if overwrite:
                os.remove(filepath)
            else:
                logger.warning(
                    "File '%s' already exists, skipping...", filepath
                )

    return inputs


def _parse_url(url):
    result = urlparse(url, allow_fragments=False)
    return result.netloc, result.path.lstrip("/")


def _single_thread_download(inputs):
    with fou.ProgressBar(total=len(inputs), iters_str="files") as pb:
        for bucket_name, obj, filepath, s3_client in pb(inputs):
            s3_client.download_file(bucket_name, obj, filepath)


def _multi_thread_download(inputs, num_workers):
    with fou.ProgressBar(total=len(inputs), iters_str="files") as pb:
        with multiprocessing.dummy.Pool(num_workers) as pool:
            for _ in pool.imap_unordered(_do_s3_download, inputs):
                pb.update()


def _do_s3_download(args):
    bucket_name, obj, filepath, s3_client = args
    s3_client.download_file(bucket_name, obj, filepath)



"""
FiftyOne v0.15.1 admin revision.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging
import json
import os


logger = logging.getLogger(__name__)


def up(db):
    pass


def down(db):
    database_name = _get_database_name()
    db = db[database_name]

    try:
        d = db["config"].find_one({})

        d.pop("type", None)

        db["config"].replace_one({"_id": d["_id"]}, d)
    except:
        pass


def _get_database_name():
    if "FIFTYONE_DATABASE_NAME" in os.environ:
        return os.environ["FIFTYONE_DATABASE_NAME"]

    config = _load_config()

    return config.get("database_name", "fiftyone")


def _load_config():
    if "FIFTYONE_CONFIG_PATH" in os.environ:
        config_path = os.environ["FIFTYONE_CONFIG_PATH"]
    else:
        config_path = os.path.join(
            os.path.expanduser("~"), ".fiftyone", "config.json"
        )

    config = {}

    if os.path.isfile(config_path):
        try:
            with open(config_path, "r") as f:
                config = json.load(f)
        except Exception as e:
            logger.warning(
                "Failed to read config from '%s': %s", config_path, str(e)
            )

    return config
